#include "launcher.h"
//#include "irrKlang.h"

#include <glm/gtx/string_cast.hpp>
#include "util.h"
#define SHADER "../resource/shaders/"
#define TEX "../resource/textures/"
static const GLfloat quad_vertices[] = {
	-0.5f, -0.5f, 0.0f,
	0.5f, -0.5f, 0.0f,
	-0.5f, 0.5f, 0.0f,
	0.5f, 0.5f, 0.0f,
};
//using namespace irrklang;
//ISoundEngine* soundEngine = createIrrKlangDevice();

float getRandomNumber(float min, float max);
glm::vec3 fadeColor(glm::vec3 startColor, glm::vec3 destColor, float ratio);
glm::vec4 getRandomBrightColor();

auto gravityForce = glm::vec3(0.0f, -GRAVITY, 0.0f);
auto noForce = glm::vec3(1.0f, 1.0f, 1.0f);

Launcher::Launcher()
	: Launcher(glm::vec3(0.0f, 0.0f, 0.0f))
{}

Launcher::Launcher(glm::vec3 position)
	: position(position)
{
	for (int i = 0; i < maxParticles; i++) 
	{
		// Initializing default particle values
		particles[i].life = -1.0f;
		particles[i].cameraDst = -1.0f;
		particles[i].trailTime = trailDelay;
		particles[i].type = Particle::Type::DEAD;
	}
}

unsigned int Launcher::particlesCount = 0;

void Launcher::Init(Camera* camera, glm::mat4& projection)
{

	shader = new Shader(SHADER"particle.vs", SHADER"particle.fs");
	this->camera = camera;
	textureId = Utils::LoadTexture(TEX"circle.bmp");

	this->particle_position = new GLfloat[maxParticles * 4];
	this->particle_color = new GLubyte[maxParticles * 4];

	glGenVertexArrays(1, &VAO);
	glBindVertexArray(VAO);


	glGenBuffers(1, &billboard_vertex_buffer);
	glBindBuffer(GL_ARRAY_BUFFER, billboard_vertex_buffer);
	glBufferData(GL_ARRAY_BUFFER, sizeof(quad_vertices), quad_vertices, GL_STATIC_DRAW);

	glGenBuffers(1, &particles_position_buffer);
	glBindBuffer(GL_ARRAY_BUFFER, particles_position_buffer);
	glBufferData(GL_ARRAY_BUFFER, maxParticles * 4 * sizeof(GLfloat), NULL, GL_STREAM_DRAW);

	glGenBuffers(1, &particles_color_buffer);
	glBindBuffer(GL_ARRAY_BUFFER, particles_color_buffer);
	glBufferData(GL_ARRAY_BUFFER, maxParticles * 4 * sizeof(GLubyte), NULL, GL_STREAM_DRAW);

	glBindVertexArray(0);

	shader->use();
	shader->setMat4("projection", projection);
	shader->setInt("sampler", 0);
}

void Launcher::Draw(bool reflect)
{
	if (!Launch) return;
	glBindVertexArray(VAO);

	// Enable additive mixing mode
	//glDisable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE); // additive mixing
	glEnable(GL_DEPTH_TEST);			//	 depth test
	glDepthMask(GL_TRUE);            //	 depth writing



	// Updating particle position buffer
	glBindBuffer(GL_ARRAY_BUFFER, particles_position_buffer);
	glBufferData(GL_ARRAY_BUFFER, maxParticles * 4 * sizeof(GLfloat), NULL, GL_STREAM_DRAW);
	glBufferSubData(GL_ARRAY_BUFFER, 0, Launcher::particlesCount * sizeof(GLfloat) * 4, particle_position);

	// Updating particle color buffer
	glBindBuffer(GL_ARRAY_BUFFER, particles_color_buffer);
	glBufferData(GL_ARRAY_BUFFER, maxParticles * 4 * sizeof(GLubyte), NULL, GL_STREAM_DRAW);
	glBufferSubData(GL_ARRAY_BUFFER, 0, Launcher::particlesCount * sizeof(GLubyte) * 4, particle_color);

	// Texture
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, textureId);


	//glEnable(GL_BLEND);
	shader->use();
	// Shader uniforms
	//glUniform3f(cameraRightId, viewMatrix[0][0], viewMatrix[1][0], viewMatrix[2][0]);
	//glUniform3f(cameraUpId, viewMatrix[0][1], viewMatrix[1][1], viewMatrix[2][1]);
	//glUniformMatrix4fv(viewProjId, 1, GL_FALSE, &vp[0][0]);
	glm::mat4 view;
	if (!reflect)
		view = glm::mat4(camera->GetViewMatrix());
	else
		view = glm::mat4(camera->GetReflectionMatrix(2));

	shader->setVec3("cameraRight", glm::vec3(view[0][0], view[1][0], view[2][0]));
	shader->setVec3("cameraUp", glm::vec3(view[0][1], view[1][1], view[2][1]));
	shader->setMat4("view", view);

	// Quad vertices
	glEnableVertexAttribArray(0);
	glBindBuffer(GL_ARRAY_BUFFER, billboard_vertex_buffer);
	glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, (void*)0);

	// Particle position
	glEnableVertexAttribArray(1);
	glBindBuffer(GL_ARRAY_BUFFER, particles_position_buffer);
	glVertexAttribPointer(1, 4, GL_FLOAT, GL_FALSE, 0, (void*)0);

	// Particle color
	glEnableVertexAttribArray(2);
	glBindBuffer(GL_ARRAY_BUFFER, particles_color_buffer);
	glVertexAttribPointer(2, 4, GL_UNSIGNED_BYTE, GL_FALSE, 0, (void*)0);

	glVertexAttribDivisor(0, 0); // Quad vertices (4 for each particle)
	glVertexAttribDivisor(1, 1); // 1 position for each quad
	glVertexAttribDivisor(2, 1); // 1 color for each quad

	glDrawArraysInstanced(GL_TRIANGLE_STRIP, 0, 4, Launcher::particlesCount);

	glDisableVertexAttribArray(0);
	glDisableVertexAttribArray(1);
	glDisableVertexAttribArray(2);

	glBindVertexArray(0);

	// Restore depth test and blending state
	//glEnable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
	//glDepthMask(GL_TRUE);

}

void Launcher::Simulate(float deltaTime)
{
	//float deltaTime = Camera::getDeltaTime();
	particlesCount = 0;

	for (int i = 0; i < maxParticles; i++)
	{
		// Simulating each particule
		Particle& p = particles[i];

		if (p.life > 0.0f)
		{
			// Alive particle
			p.speed += gravityForce * deltaTime;
			p.pos += p.speed * deltaTime;
			p.cameraDst = glm::distance(p.pos, camera->Position);

			RenderTrails(p, deltaTime);

			// Fill the GPU buffer
			particle_position[4 * particlesCount + 0] = p.pos.x;
			particle_position[4 * particlesCount + 1] = p.pos.y;
			particle_position[4 * particlesCount + 2] = p.pos.z;

			float lifeRatio = 1.0f;
			auto newColor = glm::vec3(p.r, p.g, p.b);
			switch (p.type)
			{
				case Particle::Type::SPARKLE:
					lifeRatio = p.life / sparkleLife;
					break;
				case Particle::Type::TRAIL:
					lifeRatio = p.life / trailLife;
					newColor = fadeColor(newColor, trailFade, lifeRatio);
					break;
			}
			particle_position[4 * particlesCount + 3] = p.size * lifeRatio;
			//std::cout << "omgomg newColor = " << newColor.r << std::endl;
			particle_color[4 * particlesCount + 0] = newColor.r;
			particle_color[4 * particlesCount + 1] = newColor.g;
			particle_color[4 * particlesCount + 2] = newColor.b;
			particle_color[4 * particlesCount + 3] = 255 * (p.life);

			particlesCount++;
			p.life -= deltaTime;
			continue;
		}

		// Particle just died
		if (p.type == Particle::Type::LAUNCHING)
			Explode(p);

		p.type = Particle::Type::DEAD;
		p.cameraDst = -1.0f;
	}
}



void Launcher::RenderTrails(Particle& p, float deltaTime)
{
	if (p.trailTime <= 0 && (p.type == Particle::Type::SPARKLE || p.type == Particle::Type::LAUNCHING))
	{

		SpawnParticle(
			p.pos,
			glm::vec3(0.0f),
			glm::vec4(p.r, p.g, p.b, p.a),
			trailSize,
			trailLife,
			Particle::Type::TRAIL
		);
		p.trailTime = trailDelay;
	}
	p.trailTime -= deltaTime;
}

void Launcher::SpawnParticle(glm::vec3 position, glm::vec3 speed, glm::vec4 color, float size, float life, Particle::Type type)
{
	int idx = FindUnusedParticle();

	particles[idx].pos = position;
	particles[idx].speed = speed;
	particles[idx].r = color.r;
	particles[idx].g = color.g;
	particles[idx].b = color.b;
	particles[idx].a = color.a;
	particles[idx].size = size;
	particles[idx].life = life;
	particles[idx].type = type;
}

void Launcher::Explode(Particle &p)
{
	int randomSound = getRandomNumber(1, 6);
	//soundEngine->play2D(explosionSounds[randomSound - 1]);

	float randSize = getRandomNumber(0, explosionSpread);
	for (int i = 0; i < sparklesPerExplosion; i++)
	{
		float randX = getRandomNumber(-1, 2);
		float randY = getRandomNumber(-1, 2);
		float randZ = getRandomNumber(-1, 2);
		float randSpread = getRandomNumber(0, explosionSpread);
		float randLife = getRandomNumber(0, 1.5f);

		SpawnParticle(
			p.pos,
			glm::normalize(glm::vec3(randX, randY, randZ)) * (explosionSize - randSpread + randSize),
			glm::vec4(p.r, p.g, p.b, p.a),
			sparkleSize,
			sparkleLife + randLife,
			Particle::Type::SPARKLE
		);
	}
}

void Launcher::LaunchFirework()
{
	float randX = getRandomNumber(-launchSpread, launchSpread);
	float randZ = getRandomNumber(-launchSpread, launchSpread);
	float rand = getRandomNumber(-150.0f, 150.0f);

	SpawnParticle(
		position + glm::vec3(rand, 0.0, rand),
		glm::vec3(randX, getRandomNumber(launchSpeed, launchSpeed + 25), randZ),
		getRandomBrightColor(),
		rocketSize,
		rocketLife,
		Particle::Type::LAUNCHING
	);
}

void Launcher::Update(float deltaTime)
{
	//float deltaTime = Camera::getDeltaTime();
	if (!Launch) return;
	launchTime -= deltaTime;
	if (launchTime <= 0)
	{
		LaunchFirework();
		launchTime = launchDelay;
	}

	//fountainTime -= deltaTime;
	//if (fountainTime <= 0) 
	//{
	//	LaunchFountain();
	//	fountainTime = fountainDelay;
	//}

	Simulate(deltaTime);
	SortParticles();
}

void Launcher::SortParticles() {
	std::sort(&particles[0], &particles[maxParticles]);
}

int Launcher::FindUnusedParticle() {

	for (int i = lastUsedId; i < maxParticles; i++) {
		if (particles[i].life < 0) {
			lastUsedId = i;
			return i;
		}
	}

	for (int i = 0; i < lastUsedId; i++) {
		if (particles[i].life < 0) {
			lastUsedId = i;
			return i;
		}
	}

	return 0;
}

glm::vec3 fadeColor(glm::vec3 startColor, glm::vec3 destColor, float ratio)
{
	return { 
		(ratio * startColor.r + (1.0f - ratio) * destColor.r) / 1.0f,
		(ratio * startColor.g + (1.0f - ratio) * destColor.g) / 1.0f,
		(ratio * startColor.b + (1.0f - ratio) * destColor.b) / 1.0f 
	};
}

float getRandomNumber(float min, float max)
{
	return ((float(rand()) / float(RAND_MAX)) * (max - min)) + min;
}

glm::vec4 getRandomBrightColor()
{
	auto rgb = glm::vec3(0.0f, 0.0f, 0.0f);
	unsigned char region, remainder, p, q, t;

	float h = getRandomNumber(0, 256);
	unsigned char s = 255;
	unsigned char v = 255;

	region = h / 43;
	remainder = (h - (region * 43)) * 6;

	p = (v * (255 - s)) >> 8;
	q = (v * (255 - ((s * remainder) >> 8))) >> 8;
	t = (v * (255 - ((s * (255 - remainder)) >> 8))) >> 8;

	switch (region)
	{
	case 0:
		rgb.r = v; rgb.g = t; rgb.b = p;
		break;
	case 1:
		rgb.r = q; rgb.g = v; rgb.b = p;
		break;
	case 2:
		rgb.r = p; rgb.g = v; rgb.b = t;
		break;
	case 3:
		rgb.r = p; rgb.g = q; rgb.b = v;
		break;
	case 4:
		rgb.r = t; rgb.g = p; rgb.b = v;
		break;
	default:
		rgb.r = v; rgb.g = p; rgb.b = q;
		break;
	}

	return glm::vec4(rgb, 1.0f);
}