#include <string>

#include "light.h"
#include "shader.h"

/* Light Class Implamentation */
Light::Light(glm::vec3 ambient/* = glm::vec3(0.1f, 0.1f, 0.1f)*/, glm::vec3 diffuse/* = glm::vec3(0.8f, 0.8f, 0.8f)*/, glm::vec3 specular/* = glm::vec3(1.0f, 1.0f, 1.0f)*/)
{
	this->ambient = ambient;
	this->diffuse = diffuse;
	this->specular = specular;
}

void Light::SetAmbient(float x, float y, float z)
{
	this->ambient.x = x;
	this->ambient.y = y;
	this->ambient.z = z;
}

void Light::SetDiffuse(float x, float y, float z)
{
	this->diffuse.x = x;
	this->diffuse.y = y;
	this->diffuse.z = z;
}

void Light::SetSpecular(float x, float y, float z)
{
	this->specular.x = x;
	this->specular.y = y;
	this->specular.z = z;
}

/* DirLight Class Implementation */
DirLight::DirLight(glm::vec3 direction/* = glm::vec3(0.0f, 0.0f, -1.0f)*/, glm::vec3 ambient/* = glm::vec3(0.1f, 0.1f, 0.1f)*/, glm::vec3 diffuse/* = glm::vec3(0.8f, 0.8f, 0.8f)*/, glm::vec3 specular/* = glm::vec3(1.0f, 1.0f, 1.0f)*/)
	: Light(ambient, diffuse, specular)
{
	this->direction = direction;
}

void DirLight::SetDirection(float x, float y, float z)
{
	this->direction.x = x;
	this->direction.y = y;
	this->direction.z = z;
}

void DirLight::SetupInShader(Shader& shader, const char* uniName)
{
	shader.use();
	std::string name(uniName);
	shader.setVec3(name + ".direction", direction.x, direction.y, direction.z);
	shader.setVec3(name + ".ambient", ambient.x, ambient.y, ambient.z);
	shader.setVec3(name + ".diffuse", diffuse.x, diffuse.y, diffuse.z);
	shader.setVec3(name + ".specular", specular.x, specular.y, specular.z);
}

/* PointLight Class Implementation */
PointLight::PointLight(glm::vec3 position, float constant/* = 1.0f*/, float linear/* = 0.09f*/, float quadratic/* = 0.0032f*/, glm::vec3 ambient/* = glm::vec3(0.1f, 0.1f, 0.1f)*/, glm::vec3 diffuse/* = glm::vec3(0.8f, 0.8f, 0.8f)*/, glm::vec3 specular/* = glm::vec3(1.0f, 1.0f, 1.0f)*/)
	: Light(ambient, diffuse, specular)
{
	this->position = position;
	this->constant = constant;
	this->linear = linear;
	this->quadratic = quadratic;
}

void PointLight::SetPosition(float x, float y, float z)
{
	this->position.x = x;
	this->position.y = y;
	this->position.z = z;
}

void PointLight::SetConstant(float constant)
{
	this->constant = constant;
}

void PointLight::SetLinear(float linear)
{
	this->linear = linear;
}

void PointLight::SetQuadratic(float quadratic)
{
	this->quadratic = quadratic;
}

void PointLight::SetupInShader(Shader& shader, const char* uniName)
{
	shader.use();
	std::string name(uniName);
	shader.setVec3(name + ".position", position.x, position.y, position.z);
	shader.setVec3(name + ".ambient", ambient.x, ambient.y, ambient.z);
	shader.setVec3(name + ".diffuse", diffuse.x, diffuse.y, diffuse.z);
	shader.setVec3(name + ".specular", specular.x, specular.y, specular.z);
	shader.setFloat(name + ".constant", constant);
	shader.setFloat(name + ".linear", linear);
	shader.setFloat(name + ".quadratic", quadratic);
}

/* SpotLight Class Implementation */
SpotLight::SpotLight(glm::vec3 position, glm::vec3 direction, float constant/* = 1.0f*/, float linear/* = 0.09f*/, float quadratic/* = 0.0032f*/, float innerCutoff/* = 0.218166f*//*12.5 deg*/, float outerCutoff/* = 0.305433f*//*17.5 deg*/, glm::vec3 ambient/* = glm::vec3(0.1f, 0.1f, 0.1f)*/, glm::vec3 diffuse/* = glm::vec3(0.8f, 0.8f, 0.8f)*/, glm::vec3 specular/* = glm::vec3(1.0f, 1.0f, 1.0f)*/)
	: Light(ambient, diffuse, specular)
{
	this->position = position;
	this->direction = direction;
	this->constant = constant;
	this->linear = linear;
	this->quadratic = quadratic;
	this->innerCutoff = innerCutoff;
	this->outerCutoff = outerCutoff;
}

void SpotLight::SetPosition(float x, float y, float z)
{
	this->position.x = x;
	this->position.y = y;
	this->position.z = z;
}

void SpotLight::SetDirection(float x, float y, float z)
{
	this->direction.x = x;
	this->direction.y = y;
	this->direction.z = z;
}

void SpotLight::SetConstant(float constant)
{
	this->constant = constant;
}

void SpotLight::SetLinear(float linear)
{
	this->linear = linear;
}

void SpotLight::SetQuadratic(float quadratic)
{
	this->quadratic = quadratic;
}

void SpotLight::SetInnerCutoff(float innerCutoff)
{
	this->innerCutoff = innerCutoff;
}

void SpotLight::SetOuterCutoff(float outerCutoff)
{
	this->outerCutoff = outerCutoff;
}

void SpotLight::SetupInShader(Shader& shader, const char* uniName)
{
	shader.use();
	std::string name(uniName);
	shader.setVec3(name + ".position", position.x, position.y, position.z);
	shader.setVec3(name + ".direction", direction.x, direction.y, direction.z);
	shader.setVec3(name + ".ambient", ambient.x, ambient.y, ambient.z);
	shader.setVec3(name + ".diffuse", diffuse.x, diffuse.y, diffuse.z);
	shader.setVec3(name + ".specular", specular.x, specular.y, specular.z);
	shader.setFloat(name + ".constant", constant);
	shader.setFloat(name + ".linear", linear);
	shader.setFloat(name + ".quadratic", quadratic);
	shader.setFloat(name + ".innerCutoff", innerCutoff);
	shader.setFloat(name + ".outerCutoff", outerCutoff);
}