#pragma once

#include <glad/glad.h>
#include <GLFW/glfw3.h>
#include <iostream>
#include <vector>
#include <glm/glm.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <glm/gtc/type_ptr.hpp>
#include <random>
#include <algorithm>
#include "shader.h"
#define SHADER "../resource/shaders/"
#define TEX "../resource/textures/"
class Shadow {
public:
    Shadow() {
    }

    void Bind() {
        glViewport(0, 0, _width, _height);
        glBindFramebuffer(GL_FRAMEBUFFER, _fbo);

    }

    void Unbind() {
        glBindFramebuffer(GL_FRAMEBUFFER, 0);
    }

    int Init(int image_width, int image_height, float camera_near, float camera_far, Camera* camera) {
        _width = image_width;
        _height = image_height;
        _cameraNear = camera_near;
        _cameraFar = camera_far;
        _camera = camera;
        shadowCascadeLevels.clear();
		shadowCascadeLevels.push_back(_cameraFar / 50.0f); //proximity level
		shadowCascadeLevels.push_back(_cameraFar / 25.0f); //mid level
		shadowCascadeLevels.push_back(_cameraFar / 10.0f); //far level
		shadowCascadeLevels.push_back(_cameraFar / 2.0f); //far level
        //shadowCascadeLevels.push_back(_cameraNear + (_cameraFar - _cameraNear) * 0.05f); // Very near level
        //shadowCascadeLevels.push_back(_cameraNear + (_cameraFar - _cameraNear) * 0.15f); // Near level
        //shadowCascadeLevels.push_back(_cameraNear + (_cameraFar - _cameraNear) * 0.4f);  // Mid level
        //shadowCascadeLevels.push_back(_cameraNear + (_cameraFar - _cameraNear) * 0.9f);  // Far level
        //shadowCascadeLevels.push_back(_cameraNear + (_cameraFar - _cameraNear) * 0.05f);
        //shadowCascadeLevels.push_back(_cameraNear + (_cameraFar - _cameraNear) * 0.2f);
        //shadowCascadeLevels.push_back(_cameraNear + (_cameraFar - _cameraNear) * 0.5f);
        //shadowCascadeLevels.push_back(_cameraFar);



        //shader = new Shader(SHADER"shadow_mapping.vs", SHADER"shadow_mapping.fs");
        //depthShader = new Shader(SHADER"shadow_mapping_depth.vs", SHADER"shadow_mapping_depth.fs", SHADER"shadow_mapping_depth.gs");

        // configure light FBO
        // -----------------------
        glGenFramebuffers(1, &_fbo);

        glGenTextures(1, &_color_tex_single);
        glBindTexture(GL_TEXTURE_2D_ARRAY, _color_tex_single);
        glTexImage3D(
            GL_TEXTURE_2D_ARRAY, 0, GL_DEPTH_COMPONENT32F, _width, _height, int(shadowCascadeLevels.size()) + 1,
            0, GL_DEPTH_COMPONENT, GL_FLOAT, nullptr);

        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER);
        glTexParameteri(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);

        constexpr float bordercolor[] = { 1.0f, 1.0f, 1.0f, 1.0f };
        glTexParameterfv(GL_TEXTURE_2D_ARRAY, GL_TEXTURE_BORDER_COLOR, bordercolor);

        glBindFramebuffer(GL_FRAMEBUFFER, _fbo);
        glFramebufferTexture(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, _color_tex_single, 0);
        glDrawBuffer(GL_NONE);
        glReadBuffer(GL_NONE);

        int status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
        if (status != GL_FRAMEBUFFER_COMPLETE)
        {
            std::cout << "ERROR::FRAMEBUFFER:: Framebuffer is not complete!";
            throw 0;
        }

        glBindFramebuffer(GL_FRAMEBUFFER, 0);

        // configure UBO
        // --------------------

        glGenBuffers(1, &matricesUBO);
        glBindBuffer(GL_UNIFORM_BUFFER, matricesUBO);
        glBufferData(GL_UNIFORM_BUFFER, sizeof(glm::mat4x4) * 16, nullptr, GL_STATIC_DRAW);
        glBindBufferBase(GL_UNIFORM_BUFFER, 0, matricesUBO);
        glBindBuffer(GL_UNIFORM_BUFFER, 0);

        // shader configuration
        // --------------------
        //shader->use();
        //shader->setInt("diffuseTexture", 0);
        //shader->setInt("shadowMap", 1);
     }

    void SetupUBO(glm::vec3 lightDir) {
        _lightDir = lightDir;
        // 0. UBO setup
        const auto lightMatrices = getLightSpaceMatrices();
        glBindBuffer(GL_UNIFORM_BUFFER, matricesUBO);
        for (size_t i = 0; i < lightMatrices.size(); ++i)
        {
            glBufferSubData(GL_UNIFORM_BUFFER, i * sizeof(glm::mat4x4), sizeof(glm::mat4x4), &lightMatrices[i]);
        }
        glBindBuffer(GL_UNIFORM_BUFFER, 0);

    }

    GLuint GetTexture() {
        return _color_tex_single;
    }

    std::vector<float>& GetShadowCascadeLevels() { return shadowCascadeLevels; }

    float GetCameraFarPlane() { return _cameraFar;  }

    void Cleanup() {
        glDeleteTextures(1, &_color_tex_single);
        
        glDeleteRenderbuffers(1, &_depth_rb_single);
       
        glBindFramebuffer(GL_FRAMEBUFFER, 0 /*UNBIND*/);
        glDeleteFramebuffers(1, &_fbo);
    }
private:
    std::vector<glm::mat4> getLightSpaceMatrices()
    {
        std::vector<glm::mat4> ret;
        for (size_t i = 0; i < shadowCascadeLevels.size() + 1; ++i)
        {
            if (i == 0)
            {
                ret.push_back(getLightSpaceMatrix(_cameraNear, shadowCascadeLevels[i]));
            }
            else if (i < shadowCascadeLevels.size())
            {
                ret.push_back(getLightSpaceMatrix(shadowCascadeLevels[i - 1], shadowCascadeLevels[i]));
            }
            else
            {
                ret.push_back(getLightSpaceMatrix(shadowCascadeLevels[i - 1], _cameraFar));
            }
        }
        return ret;
    }

    glm::mat4 getLightSpaceMatrix(const float nearPlane, const float farPlane)
    {
        glm::mat4 proj = glm::perspective(
            glm::radians(_camera->Zoom), (float)_width / (float)_height, nearPlane,
            farPlane);
        std::vector<glm::vec4> corners = getFrustumCornersWorldSpace(proj, _camera->GetViewMatrix());

        glm::vec3 center = glm::vec3(0, 0, 0);
        for (const auto& v : corners)
        {
            center += glm::vec3(v);
        }
        center /= corners.size();

        const auto lightView = glm::lookAt(center + _lightDir, center, glm::vec3(0.0f, 1.0f, 0.0f));

        float minX = std::numeric_limits<float>::max();
        float maxX = std::numeric_limits<float>::lowest();
        float minY = std::numeric_limits<float>::max();
        float maxY = std::numeric_limits<float>::lowest();
        float minZ = std::numeric_limits<float>::max();
        float maxZ = std::numeric_limits<float>::lowest();
        for (const auto& v : corners)
        {
            const auto trf = lightView * v;
            minX = std::min(minX, trf.x);
            maxX = std::max(maxX, trf.x);
            minY = std::min(minY, trf.y);
            maxY = std::max(maxY, trf.y);
            minZ = std::min(minZ, trf.z);
            maxZ = std::max(maxZ, trf.z);
        }

        // Tune this parameter according to the scene
        constexpr float zMult = 10.0f;
        if (minZ < 0)
        {
            minZ *= zMult;
        }
        else
        {
            minZ /= zMult;
        }
        if (maxZ < 0)
        {
            maxZ /= zMult;
        }
        else
        {
            maxZ *= zMult;
        }

        const glm::mat4 lightProjection = glm::ortho(minX, maxX, minY, maxY, minZ, maxZ);
        return lightProjection * lightView;
    }

    std::vector<glm::vec4> getFrustumCornersWorldSpace(const glm::mat4& projview)
    {
        const auto inv = glm::inverse(projview);

        std::vector<glm::vec4> frustumCorners;
        for (unsigned int x = 0; x < 2; ++x)
        {
            for (unsigned int y = 0; y < 2; ++y)
            {
                for (unsigned int z = 0; z < 2; ++z)
                {
                    const glm::vec4 pt = inv * glm::vec4(2.0f * x - 1.0f, 2.0f * y - 1.0f, 2.0f * z - 1.0f, 1.0f);
                    frustumCorners.push_back(pt / pt.w);
                }
            }
        }

        return frustumCorners;
    }

    std::vector<glm::vec4> getFrustumCornersWorldSpace(const glm::mat4& proj, const glm::mat4& view)
    {
        return getFrustumCornersWorldSpace(proj * view);
    }


protected:
    //Shader* shader;
    //Shader* depthShader;
    int _width;
    int _height;
    GLuint _fbo;

    GLuint _depth_rb_single;

    GLuint _color_tex_single;
    GLuint matricesUBO;

    glm::vec3 _lightDir;
    float _cameraNear;
    float _cameraFar;
    Camera* _camera;

    std::vector<float> shadowCascadeLevels;
    int debugLayer = 0;
};
