#pragma once

#include <glad/glad.h>
#include <iostream>
#include <glm/glm.hpp>
#include <vector>
#include "shader.h"
#include "util.h"
#include "camera.h"
#define SHADER "../resource/shaders/"
#define TEX "../resource/textures/"
#define OBJ "../resource/objects/"
class Skybox {
public:
    Skybox() {

    }

    void Init(Camera* camera, glm::mat4& projection)
    {
        m_camera = camera;
        m_shader = new Shader(SHADER"skybox.vs", SHADER"skybox.fs");
        std::vector<std::string> faces
        {
            TEX"/skybox/vz_sinister_ocean_right.png",
            TEX"/skybox/vz_sinister_ocean_left.png",
            TEX"/skybox/vz_sinister_ocean_up.png",
            TEX"/skybox/vz_sinister_ocean_down.png",
            TEX"/skybox/vz_sinister_ocean_front.png",
            TEX"/skybox/vz_sinister_ocean_back.png"
        };
        m_texture = Utils::LoadCubemap(faces);
        m_vao = Utils::InitSkyboxCube();
        m_shader->use();
        m_shader->setMat4("projection", projection);
        m_shader->setInt("skybox", 0);
    }

    void Draw(bool reflect)
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, m_texture);
        // draw skybox as last
        glDepthFunc(GL_LEQUAL);  // change depth function so depth test passes when values are equal to depth buffer's content
        m_shader->use();
        glm::mat4 view;
        if (!reflect) 
            view = glm::mat4(glm::mat3(m_camera->GetViewMatrix())); // remove translation from the view matrix
        else
            view = glm::mat4(glm::mat3(m_camera->GetReflectionMatrix(2))); // remove translation from the view matrix
        m_shader->setMat4("view", view);
        Utils::RenderCube(m_vao);
        glDepthFunc(GL_LESS); // set depth function back to default
    }

protected:
    Shader* m_shader;
    Camera* m_camera;
    GLuint m_texture;
    GLuint m_vao;

};
