#pragma once
#include <cmath>
#include <vector>
#include <glm\glm.hpp>
#define PI 3.141592653
class Sphere
{
private:
	int numVertices;
	int numIndices;
	std::vector<int> indices;
	std::vector<glm::vec3> vertices;
	std::vector<glm::vec2> texCoords;
	std::vector<glm::vec3> normals;
	std::vector<glm::vec3> tangents;
	void init(int prec) {
		numVertices = (prec + 1) * (prec + 1);
		numIndices = prec * prec * 6;
		for (int i = 0; i < numVertices; i++) { vertices.push_back(glm::vec3()); }
		for (int i = 0; i < numVertices; i++) { texCoords.push_back(glm::vec2()); }
		for (int i = 0; i < numVertices; i++) { normals.push_back(glm::vec3()); }
		for (int i = 0; i < numVertices; i++) { tangents.push_back(glm::vec3()); }
		for (int i = 0; i < numIndices; i++) { indices.push_back(0); }

		// calculate triangle vertices
		for (int i = 0; i <= prec; i++) {
			for (int j = 0; j <= prec; j++) {
				float y = (float)cos(radians(180.0f - i * 180.0f / prec));
				float x = -(float)cos(radians(j * 360.0f / prec)) * (float)abs(cos(asin(y)));
				float z = (float)sin(radians(j * 360.0f / (float)(prec))) * (float)abs(cos(asin(y)));

				float xSegment = (float)x / (float)prec;
				float ySegment = (float)y / (float)prec;
				float xPos = std::cos(xSegment * 2.0f * PI) * std::sin(ySegment * PI);
				float yPos = std::cos(ySegment * PI);
				float zPos = std::sin(xSegment * 2.0f * PI) * std::sin(ySegment * PI);
				//vertices[i * (prec + 1) + j] = glm::vec3(x, y, z);
				vertices.push_back(glm::vec3(xPos, yPos, zPos));
				texCoords[i * (prec + 1) + j] = glm::vec2(((float)j / prec), ((float)i / prec));
				normals[i * (prec + 1) + j] = glm::vec3(x, y, z);

				// calculate tangent vector
				if (((x == 0) && (y == 1) && (z == 0)) || ((x == 0) && (y == -1) && (z == 0))) {
					tangents[i * (prec + 1) + j] = glm::vec3(0.0f, 0.0f, -1.0f);
				}
				else {
					tangents[i * (prec + 1) + j] = glm::cross(glm::vec3(0.0f, 1.0f, 0.0f), glm::vec3(x, y, z));
				}
			}
		}
		// calculate triangle indices
		for (int i = 0; i < prec; i++) {
			for (int j = 0; j < prec; j++) {
				indices[6 * (i * prec + j) + 0] = i * (prec + 1) + j;
				indices[6 * (i * prec + j) + 1] = i * (prec + 1) + j + 1;
				indices[6 * (i * prec + j) + 2] = (i + 1) * (prec + 1) + j;
				indices[6 * (i * prec + j) + 3] = i * (prec + 1) + j + 1;
				indices[6 * (i * prec + j) + 4] = (i + 1) * (prec + 1) + j + 1;
				indices[6 * (i * prec + j) + 5] = (i + 1) * (prec + 1) + j;
			}
		}
	}
	float radians(float degrees) { return (degrees * 2.0f * PI) / 360.0f; }

public:
	Sphere() { init(48); };
	Sphere(int prec) { init(prec); };
	int getNumVertices() { return numVertices; }
	int getNumIndices() { return numIndices; }
	std::vector<int> getIndices() { return indices; }
	std::vector<glm::vec3> getVertices() { return vertices; }
	std::vector<glm::vec2> getTexCoords() { return texCoords; }
	std::vector<glm::vec3> getNormals() { return normals; }
	std::vector<glm::vec3> getTangents() { return tangents; }
};